/*******************************************************************************

  Copyright(c) 2007 Intel Corporation.

  This program is free software; you can redistribute it and/or modify it
  under the terms and conditions of the GNU General Public License,
  version 2, as published by the Free Software Foundation.

  This program is distributed in the hope it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.

  The full GNU General Public License is included in this distribution in
  the file called "COPYING".

*******************************************************************************/
#include "e1000.h"
#include <linux/jhash.h>
#include "../../HiCuts/hicuts.h"

/* Lincoln Tunnel -SMO */
#include <linux/tcp.h>

extern int e1000_open(struct net_device *);

#define INLINE inline

dtn_node     *glbl_root_dtn_node = NULL;

/* Modified kmem_cache_t to kmem_cache for 2.6.27 -SMO */
struct kmem_cache *g_pActionMemCache = NULL;

static u32    gs_u32XferBufSize = 0;
static u8    *gs_pu8XferBuffer  = NULL;

static INLINE boolean_t IsValidIPPacket(struct sk_buff *skb)
{
    struct iphdr *ip = (struct iphdr *)skb->data;
    uint16_t      u16Len = 0;

    // Buffer must hold the minimum IP header
    if(!pskb_may_pull(skb, sizeof(struct iphdr)))
    {
    }
    // Check header len and version
    else if(ip->ihl < 5 && ip->version != 4)
    {
    }
    // Buffer must hold at least the advertized header length
    else if(!pskb_may_pull(skb, (ip->ihl * 4)))
    {
    }
    // Checksum the header
    else if(skb->ip_summed != CHECKSUM_UNNECESSARY &&
            skb->ip_summed != CHECKSUM_COMPLETE    &&
            ip_fast_csum(skb->data, ip->ihl))
    {
    }
    else
    {
        // Get the expected length
        u16Len = __constant_ntohs(ip->tot_len);

        // Make sure the length makes sense and trim the skb if it does
        if((u16Len <= skb->len) && ((ip->ihl * 4) <= u16Len) &&
           !pskb_trim_rcsum(skb, u16Len))
        {
            return TRUE;
        }
    }

    return FALSE;
}

static INLINE boolean_t IsProcessedPacket(struct sk_buff *skb,
                                          UTupple_t *puTupple)
{
    struct e1000_adapter *adapter = NULL;
    struct ethhdr        *eth = eth_hdr(skb);

    if(eth && __constant_htons(ETH_P_IP) == eth->h_proto    &&
       skb && skb->dev && (adapter = netdev_priv(skb->dev)) &&
       adapter->m_aHashLevels[0].m_apFlows && adapter->m_pThreadCtxt &&
       IsValidIPPacket(skb))
    {
        struct iphdr *ip=(struct iphdr *)skb->data;

        memset(puTupple, 0x00, sizeof(UTupple_t));

        puTupple->m_sTuppleBits.m_u32SrcAddr = ip->saddr;
        puTupple->m_sTuppleBits.m_u32DstAddr = ip->daddr;
        puTupple->m_sTuppleBits.m_u8Protocol = ip->protocol;

        switch(ip->protocol)
        {
            case IPPROTO_TCP:
                {
                    struct tcphdr *tcp =
                        (struct tcphdr *)&skb->data[ip->ihl * 4];
                    puTupple->m_sTuppleBits.m_u16SrcPort = tcp->source;
                    puTupple->m_sTuppleBits.m_u16DstPort = tcp->dest;
                    return TRUE;
                }
                break;

            case IPPROTO_UDP:
                {
                    struct udphdr *udp =
                        (struct udphdr *)&skb->data[ip->ihl * 4];
                    puTupple->m_sTuppleBits.m_u16SrcPort = udp->source;
                    puTupple->m_sTuppleBits.m_u16DstPort = udp->dest;
                    return TRUE;
                }
                break;

#ifdef FLOWS_DEBUG
            case 253:
	    case 254: // added shrikant
                {
                    puTupple->m_sTuppleBits.m_u16SrcPort = 0;
                    puTupple->m_sTuppleBits.m_u16DstPort = 0;
                    return TRUE;
                }
                break;
#endif /* FLOWS_DEBUG */

            default:
                break;
        }
    }

    return FALSE;
}

static INLINE void QForFlowLookup(struct sk_buff *skb,
                                  UTupple_t *puTupple,
                                  EQEntryType_t eType)
{
    struct e1000_adapter *adapter  = NULL;
    SThreadCtxt_t        *pContext = NULL;
    SQEntry_t            *psEntry  = NULL;
    static uint32_t       s_u32Flags = 0;
    uint32_t              u32Mask = 0;

    if(!skb || !puTupple)
    {
        return;
    }

    adapter  = (struct e1000_adapter *)netdev_priv(skb->dev);
    pContext = adapter->m_pThreadCtxt;
    psEntry  = &pContext->m_asQ[pContext->m_iWIndex];

    if(!pContext->m_pTask)
    {
        printk(KERN_WARNING "%d: Task is dead!\n", adapter->bd_number);
    }
    else if(FLOW_Q_TYPE_EMPTY != psEntry->m_eType)
    {
        u32Mask = 1 << adapter->bd_number;

        if(!(s_u32Flags & u32Mask))
        {
            printk(KERN_WARNING "%d: Q wrap!\n", adapter->bd_number);
            s_u32Flags |= u32Mask;
        }
    }
    else
    {
        memcpy(&psEntry->m_uTupple, puTupple, sizeof(UTupple_t));
        psEntry->m_pNetDev = skb->dev;
        switch(eType)
        {
            case FLOW_Q_TYPE_DEL:
                psEntry->skb = NULL;
                break;
            default:
                psEntry->skb = skb;
                break;
        }

        // The thread keys off of this so it should be the last thing filled
        psEntry->m_eType = eType;
        send_sig(SIGHUP, pContext->m_pTask, 1);
        pContext->m_iWIndex = (pContext->m_iWIndex + 1) % FLOW_Q_DEPTH;

        return;
    }

    switch(eType)
    {
        case FLOW_Q_TYPE_DEL:
            // Do nothing
            break;
        default:
            dev_kfree_skb_any(skb);
            break;
    }
}

static INLINE boolean_t FlowSearch(struct e1000_adapter *adapter,
                                   UTupple_t *puTupple,
                                   SFlow_t **ppsFlow,
                                   SHashLevel_t **ppHashLevel,
                                   u32 *pu32Index)
{
    SHashLevel_t         *pHashLevels = adapter->m_aHashLevels;
    SFlow_t              *psFlow = NULL;
    u32                   u32Index = 0;
    int                   ii;
#ifdef FLOWS_DEBUG
    cycles_t              cyBefore;
    cycles_t              cyAfter;
    cycles_t              cyDiff;
#endif /* FLOWS_DEBUG */

#ifdef FLOWS_DEBUG
    cyBefore = get_cycles();
#endif /* FLOWS_DEBUG */

    for(ii = 0;ii < FLOW_MAX_HASH_LEVELS;ii++)
    {
        // Compute hash index
        u32Index = (jhash2(puTupple->m_au32Data,
                           TUPPLE_U32_LEN, pHashLevels[ii].m_u32InitVal) %
                    pHashLevels[ii].m_iSize);

        if(ppHashLevel)
        {
            *ppHashLevel = &pHashLevels[ii];
        }

        if(pu32Index)
        {
            *pu32Index = u32Index;
        }

        // See if there is a flow
        if((psFlow = pHashLevels[ii].m_apFlows[u32Index]))
        {
#ifdef FLOWS_DEBUG
            if(ii == 0)
            {
                cyAfter = get_cycles();

                if(cyBefore < cyAfter)
                {
                    adapter->m_acySamples[adapter->m_iIndex] = cyDiff =
                        (cyAfter - cyBefore);

                    adapter->m_iIndex =
                        (adapter->m_iIndex + 1) % DEBUG_SAMPLES;

                    if(adapter->m_iNumSamples)
                    {
                        if(cyDiff < adapter->m_cyMin)
                        {
                            adapter->m_cyMin = cyDiff;
                        }

                        if(adapter->m_cyMax < cyDiff)
                        {
                            adapter->m_cyMax = cyDiff;
                        }
                    }
                    else
                    {
                        adapter->m_cyMin = cyDiff;
                        adapter->m_cyMax = cyDiff;
                    }

                    if(adapter->m_iNumSamples < DEBUG_SAMPLES)
                    {
                        adapter->m_iNumSamples++;
                    }
                }
            }
#endif /* FLOWS_DEBUG */

            // Check for collisions
            if((psFlow->m_uTupple.m_au32Data[0] != puTupple->m_au32Data[0]) ||
               (psFlow->m_uTupple.m_au32Data[1] != puTupple->m_au32Data[1]) ||
               (psFlow->m_uTupple.m_au32Data[2] != puTupple->m_au32Data[2]) ||
               (psFlow->m_uTupple.m_au32Data[3] != puTupple->m_au32Data[3]))
            {
#ifdef FLOWS_DEBUG
                if(!(FLOW_COLLISION & psFlow->m_u32Flags))
                {
                    psFlow->m_u32Flags |= FLOW_COLLISION;
                    pHashLevels[ii].m_u32Cols++;
                }
#endif /* FLOWS_DEBUG */
            }
            else
            {
                if(ppsFlow)
                {
                    *ppsFlow = psFlow;
                    prefetch(psFlow);
                }
                return TRUE;
            }
        }
        else
        {
            // There is no flow so stop searching
            break;
        }
    }

    return FALSE;
}

static INLINE void RecoupPort(struct e1000_adapter *adapter,
                              SFlow_t *psFlow)
{
    // See if this is a main NAT entry and we need to recoup the port
    // number
    if(((FLOW_FLAG_PAIR | FLOW_FLAG_NAT) & psFlow->m_u32Flags) ==
       FLOW_FLAG_NAT)
    {
        u16 u16Port = psFlow->m_uPairTupple.m_sTuppleBits.m_u16DstPort;

        if(adapter->m_u32NextPort)
        {
            adapter->m_u32NextPort--;
        }

        adapter->m_au16NatPorts[adapter->m_u32NextPort] = u16Port;
    }
}

static INLINE boolean_t FlowExists(struct sk_buff *skb,
                                   UTupple_t *puTupple,
                                   SFlow_t **ppsFlow,
                                   SHashLevel_t **ppHashLevel,
                                   u32 *pu32Index)
{
    struct e1000_adapter *adapter = netdev_priv(skb->dev);
    struct timeval        tvNow;
    SHashLevel_t         *pHashLevel = NULL;
    boolean_t             bRet = FALSE;
    SFlow_t              *psFlow = NULL;
    u32                   u32Index = 0;

#ifdef HYPER_OPT
    if(adapter->m_psFlowCache)
    {
        *ppsFlow = adapter->m_psFlowCache;
        prefetch(adapter->m_psFlowCache);
        return TRUE;
    }
#endif /* HYPER_OPT */

    if(FlowSearch(adapter, puTupple, &psFlow, &pHashLevel, &u32Index))
    {
        // Check the age - it's in order of minutes so just deal
        // with tv_sec
        do_gettimeofday(&tvNow);

        if(((tvNow.tv_sec - psFlow->m_tvLastAccess.tv_sec) / 60) <=
           FLOW_MAX_AGE_MINUTES)
        {
            psFlow->m_tvLastAccess.tv_sec = tvNow.tv_sec;

            if(ppsFlow)
            {
                *ppsFlow = psFlow;
#ifdef HYPER_OPT
                adapter->m_psFlowCache = psFlow;
#endif /* HYPER_OPT */
            }

            bRet = TRUE;
        }
        else
        {
            // Delete old flow
            pHashLevel->m_apFlows[u32Index] = NULL;

            // See if this is a paired flow
            if((FLOW_FLAG_NAT | FLOW_FLAG_LOADB) & psFlow->m_u32Flags)
            {
                struct net_device  *pOldDev = skb->dev;

                RecoupPort(adapter, psFlow);

                skb->dev = psFlow->m_pPairNetDev;
                QForFlowLookup(skb, &psFlow->m_uPairTupple,
                               FLOW_Q_TYPE_DEL);
                skb->dev = pOldDev;
            }
            kmem_cache_free(g_pActionMemCache, psFlow);
        }
    }

    if(ppHashLevel)
    {
        *ppHashLevel = pHashLevel;
    }

    if(pu32Index)
    {
        *pu32Index = u32Index;
    }

    return bRet;
}

static INLINE boolean_t CheckCondition(u32 u32Val1, u32 u32Val2, EOP_t op)
{
    if((u32Val1 == u32Val2) && ((OP_EQ  == op) ||
                                (OP_GTE == op) ||
                                (OP_LTE == op)))
    {
        return TRUE;
    }
    else if((u32Val1 != u32Val2) && (OP_EQ  == op))
    {
        return TRUE;
    }
    else if(((OP_LT == op) || (OP_LTE == op)) && (u32Val1 < u32Val2))
    {
        return TRUE;
    }
    else if(((OP_GT == op) || (OP_GTE == op)) && (u32Val1 > u32Val2))
    {
        return TRUE;
    }

    return FALSE;
}

static INLINE void ProcessFlow(struct sk_buff *skb,
                               SFlow_t *psFlow,
                               UTupple_t *puTupple,
                               boolean_t bInThread)
{
    struct e1000_adapter *adapter  = NULL;
    struct ethhdr        *eth      = NULL;
    struct iphdr         *ip       = NULL;
    unsigned int          uiIndex;
    SAction_t            *psAction = NULL;
    SOffset_t            *psOffset = NULL;
    SRule_t              *psRule = NULL;
    uint32_t              u32CO = 0;
    uint32_t              u32ModOffset;
    uint32_t              u32ModValue;
    uint32_t              u32Value;

    for(uiIndex = 0;uiIndex < FLOW_MAX_RULES_PER;)
    {
        psRule = &psFlow->m_aRules[uiIndex];

        // Perform Offset operation
        switch(psRule->m_sOffset.m_eOffsetType)
        {
            case OFFTYPE_CO_RELATIVE:
                u32CO += psRule->m_sOffset.m_u32ValOrOffset;
                break;

            case OFFTYPE_L3_ABSOLUTE:
                u32CO = psRule->m_sOffset.m_u32ValOrOffset;
                break;

            case OFFTYPE_CO_COMPUTED:
                u32CO = (((skb->data[u32CO] & psRule->m_sOffset.m_u32Mask) >>
                          psRule->m_sOffset.m_u32Shift) +
                         psRule->m_sOffset.m_u32ValOrOffset);
                break;

            case OFFTYPE_NO_CHANGE:
            default:
                break;
        }

        if(skb->len <= u32CO)
        {
            break;
        }

        if(RULE_NO_CONDITION & psRule->m_sCondition.m_u32Flags)
        {
            psAction = &(psRule->m_sTrueAction);
        }
        else
        {
            if(RULE_USE_FLOWVAR & psRule->m_sCondition.m_u32Flags)
            {
                u32Value =
                    psFlow->m_au32Values[psRule->m_sCondition.index];
            }
            else
            {
                u32Value = skb->data[u32CO];
            }

            // Compute condition and determine action
            if(CheckCondition(psRule->m_sCondition.result,
                              (u32Value &
                               psRule->m_sCondition.mask),
                              psRule->m_sCondition.op))
            {
                psAction = &(psRule->m_sTrueAction);
            }
            else
            {
                psAction = &(psRule->m_sFalseAction);
            }
        }

        // Perform Action
        switch(psAction->m_eActionType)
        {
            case FLOWACTION_PASS:
                if(!psAction->m_uAction.m_sParamPass.m_pOutDev)
                {
					/* Lincoln Tunnel -SMO
					 * Changed paramter list for dev_get_by_name to contain &init_net 
					 */
                    struct net_device *pDev =
                        dev_get_by_name(&init_net, psAction->m_uAction.m_sParamPass.m_acOutDevName);
                    if((psAction->m_uAction.m_sParamPass.m_pOutDev = pDev))
                    {
                        dev_put(pDev);

                        // Make sure this device belongs to us
                        if(pDev->open != &e1000_open)
                        {
                            psAction->m_uAction.m_sParamPass.m_pOutDev = NULL;
                        }
                        else
                        {
                            psAction->m_uAction.m_sParamPass.m_u32Addr =
                                __constant_htonl(psAction->m_uAction.m_sParamPass.m_u32Addr);
                        }
                    }
                }

                // Could not find the output device!
                if(!psAction->m_uAction.m_sParamPass.m_pOutDev)
                {
                    if(!(ACTION_PRINTED_DEV_ERROR & psAction->m_u32Flags))
                    {
                        psAction->m_u32Flags |= ACTION_PRINTED_DEV_ERROR;
                        printk(KERN_ERR "Could not find device <%s>!\n",
                               psAction->m_uAction.m_sParamPass.m_acOutDevName);
                    }
                    dev_kfree_skb_any(skb);
                    return;
                }

                skb->dev = psAction->m_uAction.m_sParamPass.m_pOutDev;
                adapter  = netdev_priv(skb->dev);
                eth      = eth_hdr(skb);
                ip       = (struct iphdr *) skb->data;

                // Update MAC info
                memcpy(eth->h_dest,
                       psAction->m_uAction.m_sParamPass.m_au8DstMac,
                       ETH_ALEN);
                memcpy(eth->h_source,adapter->hw.mac.addr, ETH_ALEN); // Lincoln Tunnel -SMO
                // Update IP info
                ip->ttl--;
                if(FLOW_FLAG_NAT & psFlow->m_u32Flags)
                {
                    uint32_t u32CS = 0;
                    uint16_t u16Len = 0;

                    if(FLOW_FLAG_PAIR & psFlow->m_u32Flags)
                    {
                        ip->daddr = psAction->m_uAction.m_sParamPass.m_u32Addr;
                    }
                    else
                    {
                        ip->saddr = psAction->m_uAction.m_sParamPass.m_u32Addr;
                    }
                    switch(ip->protocol)
                    {
                        case IPPROTO_TCP:
                            {
                                struct tcphdr *tcp =
                                    (struct tcphdr *)&skb->data[ip->ihl * 4];
                                if(FLOW_FLAG_PAIR & psFlow->m_u32Flags)
                                {
                                    tcp->dest =
                                        psFlow->m_uPairTupple.m_sTuppleBits.m_u16SrcPort;
                                }
                                else
                                {
                                    tcp->source =
                                        psFlow->m_uPairTupple.m_sTuppleBits.m_u16DstPort;
                                }

                                tcp->check = 0;
                                u16Len = skb->len - (ip->ihl * 4);
                                u32CS = csum_partial((char *)tcp, u16Len, 0);
                                tcp->check =
                                    csum_tcpudp_magic(ip->saddr, ip->daddr,
                                                      u16Len, IPPROTO_TCP,
                                                      u32CS);
                            }
                            break;

                        case IPPROTO_UDP:
                            {
                                struct udphdr *udp =
                                    (struct udphdr *)&skb->data[ip->ihl * 4];
                                if(FLOW_FLAG_PAIR & psFlow->m_u32Flags)
                                {
                                    udp->dest =
                                        psFlow->m_uPairTupple.m_sTuppleBits.m_u16SrcPort;
                                }
                                else
                                {
                                    udp->source =
                                        psFlow->m_uPairTupple.m_sTuppleBits.m_u16DstPort;
                                }

                                udp->check = 0;
                                u16Len = __constant_ntohs(udp->len);
                                u32CS = csum_partial((char *)udp, u16Len, 0);
                                udp->check =
                                    csum_tcpudp_magic(ip->saddr, ip->daddr,
                                                      u16Len, IPPROTO_UDP,
                                                      u32CS);
                            }
                            break;
                    }
                }
                else if(FLOW_FLAG_LOADB & psFlow->m_u32Flags)
                {
                    if(FLOW_FLAG_PAIR & psFlow->m_u32Flags)
                    {
                        ip->saddr =
                            psAction->m_uAction.m_sParamPass.m_u32Addr;
                    }
                    else
                    {
                        ip->daddr =
                            psAction->m_uAction.m_sParamPass.m_u32Addr;
                    }
                }

                ip->check = 0;
                ip->check = ip_fast_csum(skb->data, ip->ihl);

                // Reset SKB and send
                skb_push(skb, ETH_HLEN);
#ifndef USE_OWN_XMIT_QS
                dev_queue_xmit(skb);
#else /* USE_OWN_XMIT_QS */
                if(bInThread)
                {
                    preempt_disable();
                }

                {
                    int iCpu = smp_processor_id();
                    int iIndex = adapter->m_aiWIndex[iCpu];

                    if(NULL == adapter->m_apBuffers[(iCpu * PER_CPU_Q_LEN) + iIndex])
                    {
                        adapter->m_apBuffers[(iCpu * PER_CPU_Q_LEN) + iIndex] = skb;
                        adapter->m_aiWIndex[iCpu]=
                            (adapter->m_aiWIndex[iCpu] + 1) % PER_CPU_Q_LEN;
                    }
                    else
                    {
                        adapter->m_iDropped++;
                        dev_kfree_skb_any(skb);
                        // dev_queue_xmit(skb);
                    }
                }

                if(bInThread)
                {
                    preempt_enable();
                }
#endif /* USE_OWN_XMIT_QS */
                return;
                break;

            case FLOWACTION_MODIFY:
                psOffset = &psAction->m_uAction.m_sParamModify.m_sOffset;

                switch(psOffset->m_eOffsetType)
                {
                    case OFFTYPE_CO_RELATIVE:
                        u32ModOffset = u32CO + psOffset->m_u32ValOrOffset;
                        break;

                    case OFFTYPE_L3_ABSOLUTE:
                        u32ModOffset = psOffset->m_u32ValOrOffset;
                        break;

                    case OFFTYPE_NO_CHANGE:
                    default:
                        u32ModOffset = u32CO;
                        break;
                }

                if(ACTION_MOD_USE_GLOBAL_VALS & psAction->m_u32Flags)
                {
                    u32ModValue =
                        psFlow->m_au32Values[psAction->m_uAction.m_sParamModify.m_u32ValOrIndex];
                }
                else
                {
                    u32ModValue =
                        psAction->m_uAction.m_sParamModify.m_u32ValOrIndex;
                }

                skb->data[u32ModOffset] =
                    ((skb->data[u32ModOffset] &
                      ~(psAction->m_uAction.m_sParamModify.m_u32Mask)) |
                     u32ModValue);
                break;

            case FLOWACTION_EXCEPTION:
                QForFlowLookup(skb, puTupple, FLOW_Q_TYPE_EXCEPTION);
                return;
                break;

            case FLOWACTION_NONE:
                break;

            case FLOWACTION_CRYPTO:
                break;

            case FLOWACTION_DROP:
            default:
                dev_kfree_skb_any(skb);
                return;
                break;
        }

        // Determine jump
        uiIndex = psAction->m_uiNextRule;
    }

    // We should never get here
    printk(KERN_ERR "Fell out of flow processing loop - "
           "uiIndex = %d u32CO = %d skb->len = %d",
           uiIndex, u32CO,  skb->len);
    dev_kfree_skb_any(skb);
}

void DoIt(struct sk_buff *skb)
{
    UTupple_t uTupple;
    SFlow_t  *psFlow=NULL;

    // See if this is a valid packet and/or one we care about
    if(IsProcessedPacket(skb, &uTupple))
    {
        // Perform actions if the flow already exists (quick path)
        // else Q the packet for flow lookup (slow path)
        if(FlowExists(skb, &uTupple, &psFlow, NULL, NULL))
        {
            ProcessFlow(skb, psFlow, &uTupple, FALSE);
        }
        else
        {
            QForFlowLookup(skb, &uTupple, FLOW_Q_TYPE_NORMAL);
        }

        return;
    }

    // We didn't care about this packet so perform normal OS processing
#ifdef CONFIG_E1000_NAPI
    netif_receive_skb(skb);
#else
    netif_rx(skb);
#endif /* CONFIG_E1000_NAPI */
}

static INLINE void CreateFlowPair(struct sk_buff *skb,
                                  UTupple_t *puMainTupple,
                                  SFlow_t *psMainFlow)
{
    struct e1000_adapter *adapter = netdev_priv(skb->dev);
    struct net_device    *pOldDev = NULL;
    struct ethhdr        *eth = NULL;
    SHashLevel_t         *pHashLevel = NULL;
    unsigned int          uiIndex;
    UTupple_t             sTupple;
    SAction_t            *psAction = NULL;
    SRule_t              *psRule = NULL;
    SFlow_t              *psFlow = NULL;
    u32                   u32Index = 0;

    // Find the "pass" rule
    for(uiIndex = 0;uiIndex < FLOW_MAX_RULES_PER;)
    {
        psRule = &psMainFlow->m_aRules[uiIndex];

        if(FLOWACTION_PASS == psRule->m_sTrueAction.m_eActionType)
        {
            psAction = &psRule->m_sTrueAction;
        }
        else if(FLOWACTION_PASS == psRule->m_sFalseAction.m_eActionType)
        {
            psAction = &psRule->m_sFalseAction;
        }
        else
        {
            continue;
        }
        break;
    }

    if(!psAction)
    {
        return;
    }

    if(!psAction->m_uAction.m_sParamPass.m_pOutDev)
    {
        if((psAction->m_uAction.m_sParamPass.m_pOutDev =
            (struct net_device *) dev_get_by_name(&init_net, psAction->m_uAction.m_sParamPass.m_acOutDevName)))
        {
            dev_put(psAction->m_uAction.m_sParamPass.m_pOutDev);
            psAction->m_uAction.m_sParamPass.m_u32Addr =
                __constant_htonl(psAction->m_uAction.m_sParamPass.m_u32Addr);
        }
    }

    if(!psAction->m_uAction.m_sParamPass.m_pOutDev)
    {
        return;
    }

    // Create our tupple
    memset(&sTupple, 0x00, sizeof(UTupple_t));

    sTupple.m_sTuppleBits.m_u8Protocol =
        puMainTupple->m_sTuppleBits.m_u8Protocol;
    sTupple.m_sTuppleBits.m_u16SrcPort =
        puMainTupple->m_sTuppleBits.m_u16DstPort;

    if(FLOW_FLAG_NAT & psMainFlow->m_u32Flags)
    {   // Ports are already in NW order
        sTupple.m_sTuppleBits.m_u16DstPort =
            adapter->m_au16NatPorts[adapter->m_u32NextPort];
        adapter->m_u32NextPort =
            (adapter->m_u32NextPort + 1) % FLOW_MAX_NAT_PORTS;
        sTupple.m_sTuppleBits.m_u32DstAddr =
            psAction->m_uAction.m_sParamPass.m_u32Addr;
        sTupple.m_sTuppleBits.m_u32SrcAddr =
            puMainTupple->m_sTuppleBits.m_u32DstAddr;
    }
    else
    {
        sTupple.m_sTuppleBits.m_u16DstPort =
            puMainTupple->m_sTuppleBits.m_u16SrcPort;
        sTupple.m_sTuppleBits.m_u32DstAddr =
            puMainTupple->m_sTuppleBits.m_u32SrcAddr;
        sTupple.m_sTuppleBits.m_u32SrcAddr =
            psAction->m_uAction.m_sParamPass.m_u32Addr;
    }

    pOldDev = skb->dev;
    skb->dev = psAction->m_uAction.m_sParamPass.m_pOutDev;

    // Find a place for it and add it
    if(!FlowExists(skb, &sTupple, &psFlow, &pHashLevel, &u32Index) &&
       pHashLevel && u32Index < pHashLevel->m_iSize &&
       NULL == pHashLevel->m_apFlows[u32Index] &&
       (psFlow = (SFlow_t *) kmem_cache_alloc(g_pActionMemCache,
                                              GFP_ATOMIC)))
    {
        memcpy(&psMainFlow->m_uPairTupple, &sTupple, sizeof(UTupple_t));
        memset(psFlow,0x00,sizeof(SFlow_t));
        memcpy(&psFlow->m_uTupple, &sTupple, sizeof(UTupple_t));
        memcpy(&psFlow->m_uPairTupple, puMainTupple, sizeof(UTupple_t));
        psFlow->m_u32Flags = (psMainFlow->m_u32Flags | FLOW_FLAG_PAIR);

        psFlow->m_pPairNetDev = pOldDev;
        psMainFlow->m_pPairNetDev = skb->dev;

        psRule = &psFlow->m_aRules[0];

        psRule->m_sOffset.m_eOffsetType = OFFTYPE_NO_CHANGE;
        psRule->m_sCondition.m_u32Flags = RULE_NO_CONDITION;
        psAction = &(psRule->m_sTrueAction);

        psAction->m_eActionType = FLOWACTION_PASS;
        psAction->m_uAction.m_sParamPass.m_pOutDev = pOldDev;
        eth = eth_hdr(skb);
        memcpy(psAction->m_uAction.m_sParamPass.m_au8DstMac,
               eth->h_source, ETH_ALEN);
        if(FLOW_FLAG_NAT & psFlow->m_u32Flags)
        {
            psAction->m_uAction.m_sParamPass.m_u32Addr =
                puMainTupple->m_sTuppleBits.m_u32SrcAddr;
        }
        else
        {
            psAction->m_uAction.m_sParamPass.m_u32Addr =
                puMainTupple->m_sTuppleBits.m_u32DstAddr;
        }

        do_gettimeofday(&psFlow->m_tvLastAccess);
        pHashLevel->m_apFlows[u32Index] = psFlow;
    }

    skb->dev = pOldDev;
}

boolean_t FindFlow(struct sk_buff *skb,
                   UTupple_t *puTupple,
                   SFlow_t **ppsFlow,
                   SHashLevel_t *pHashLevel,
                   u32 u32Index,
                   SThreadCtxt_t *pContext)
{
    unsigned int  uiDimension;
    unsigned int  uiSrcIp;
    unsigned int  uiDstIp;
    unsigned int  uiSrcPort;
    unsigned int  uiDstPort;
    unsigned int  uiProtocol;
#ifdef FLOWS_DEBUG
    cycles_t      cyBefore;
    cycles_t      cyAfter;
    cycles_t      cyDiff;
#endif /* FLOWS_DEBUG */
    dtn_node     *node = glbl_root_dtn_node;
    SFlow_t      *psFlow = NULL;
    rule         *pRule = NULL;
    int           iIndex;
    int           ii;

    uiSrcIp = __constant_ntohl(puTupple->m_sTuppleBits.m_u32SrcAddr);
    uiDstIp = __constant_ntohl(puTupple->m_sTuppleBits.m_u32DstAddr);

    uiSrcPort = __constant_ntohs(puTupple->m_sTuppleBits.m_u16SrcPort);
    uiDstPort = __constant_ntohs(puTupple->m_sTuppleBits.m_u16DstPort);

    switch(puTupple->m_sTuppleBits.m_u8Protocol)
    {
        case IPPROTO_TCP:
            uiProtocol = PROT_TCP;
            break;
        case IPPROTO_UDP:
            uiProtocol = PROT_UDP;
            break;
        default:
            uiProtocol = puTupple->m_sTuppleBits.m_u8Protocol;
            break;
    }

    // See if we have the decision tree and there is room for this flow
    if(gs_pu8XferBuffer && pHashLevel && glbl_root_dtn_node &&
       u32Index < pHashLevel->m_iSize &&
       NULL == pHashLevel->m_apFlows[u32Index])
    {
#ifdef FLOWS_DEBUG
        cyBefore = get_cycles();
#endif /* FLOWS_DEBUG */

        while(node && (node->num_of_rules || node->sub_tree_size))
        {
            // Try to find the rule
            for(ii = 0;ii < node->num_of_rules;ii++)
            {
                pRule = node->r_list[ii];

                if(CHECK_FIELD(pRule, protocol, uiProtocol) &&
                   CHECK_FIELD(pRule, src_ip,   uiSrcIp)    &&
                   CHECK_FIELD(pRule, dst_ip,   uiDstIp)    &&
                   CHECK_FIELD(pRule, src_prt,  uiSrcPort)  &&
                   CHECK_FIELD(pRule, dst_prt,  uiDstPort))
                {
#ifdef FLOWS_DEBUG
                    cyAfter = get_cycles();

                    if(cyBefore < cyAfter)
                    {
                        pContext->m_acySamples[pContext->m_iSampleIndex] =
                            cyDiff = (cyAfter - cyBefore);

                        pContext->m_iSampleIndex =
                            (pContext->m_iSampleIndex + 1) % DEBUG_SAMPLES;

                        if(pContext->m_iNumSamples)
                        {
                            if(cyDiff < pContext->m_cyMin)
                            {
                                pContext->m_cyMin = cyDiff;
                            }

                            if(pContext->m_cyMax < cyDiff)
                            {
                                pContext->m_cyMax = cyDiff;
                            }
                        }
                        else
                        {
                            pContext->m_cyMin = cyDiff;
                            pContext->m_cyMax = cyDiff;
                        }

                        if(pContext->m_iNumSamples < DEBUG_SAMPLES)
                        {
                            pContext->m_iNumSamples++;
                        }
                    }
#endif /* FLOWS_DEBUG */

                    if((psFlow =
                        (SFlow_t *) kmem_cache_alloc(g_pActionMemCache,
                                                     GFP_ATOMIC)))
                    {
                        memset(psFlow,0x00,sizeof(SFlow_t));
                        if(RULE_TYPE_NAT == pRule->rule_type)
                        {
                            psFlow->m_u32Flags = FLOW_FLAG_NAT;
                        }
                        else if(RULE_TYPE_LOADB == pRule->rule_type)
                        {
                            psFlow->m_u32Flags = FLOW_FLAG_LOADB;
                        }
                        memcpy(psFlow->m_aRules, pRule->rule_array,
                               FLOW_MAX_RULES_PER * sizeof(SRule_t));
                        memcpy(psFlow->m_au32Values, pRule->global_vars,
                               FLOW_MAX_VALUES_PER * sizeof(u32));
                        memcpy(&psFlow->m_uTupple, puTupple,
                               sizeof(UTupple_t));

                        if((FLOW_FLAG_NAT | FLOW_FLAG_LOADB) &
                           psFlow->m_u32Flags)
                        {
                            CreateFlowPair(skb, puTupple, psFlow);
                        }

                        do_gettimeofday(&psFlow->m_tvLastAccess);

                        pHashLevel->m_apFlows[u32Index] = psFlow;
                        *ppsFlow = psFlow;
                        return TRUE;
                    }
                    else
                    {
                        static boolean_t s_bPrint = TRUE;
                        if(s_bPrint)
                        {
                            printk(KERN_ERR "Allocation problem!\n");
                            s_bPrint = FALSE;
                        }
                        return FALSE;
                    }
                }
            }

            if(node->sub_tree_size && node->sub_tree)
            {
                // figure out which branch to take
                switch(node->n_dim)
                {
                    case IP_SRC:
                        uiDimension = uiSrcIp;
                        break;
                    case IP_DEST:
                        uiDimension = uiDstIp;
                        break;
                    case SRC_PRT:
                        uiDimension = uiSrcPort;
                        break;
                    case DST_PRT:
                        uiDimension = uiDstPort;
                        break;
                    case PROTOCOL:
                        uiDimension = uiProtocol;
                        break;
                    default:
                        return FALSE;
                        break;
                }

                if(uiDimension < node->min_max.low ||
                   uiDimension > node->min_max.high)
                {
                    break;
                }
                else
                {
                    if((iIndex = ((uiDimension - node->min_max.low) /
                                  node->dim_slice)) == node->sub_tree_size)
                    {
                        iIndex--;
                    }

                    node = node->sub_tree[iIndex];
                }
            }
            else
            {
                // No more nodes to search
                break;
            }
        }
    }

    return FALSE;
}

void ProcessException(struct sk_buff *skb,
                      UTupple_t *puTupple)
{
    // For right now
    dev_kfree_skb_any(skb);
}

void ProccessDel(struct net_device *dev, UTupple_t *puTupple)
{
    struct e1000_adapter *adapter = netdev_priv(dev);
    SHashLevel_t         *pHashLevel = NULL;
    SFlow_t              *psFlow = NULL;
    u32                   u32Index = 0;

    if(FlowSearch(adapter, puTupple, &psFlow, &pHashLevel, &u32Index))
    {
        // Delete old flow
        pHashLevel->m_apFlows[u32Index] = NULL;

        printk(KERN_INFO "%s: Deleted %d @ %p\n",
               __FUNCTION__, u32Index, pHashLevel->m_apFlows);

        RecoupPort(adapter, psFlow);

        kmem_cache_free(g_pActionMemCache, psFlow);
    }
}

int FlowThread(void *pvArg)
{
    SThreadCtxt_t *pContext = (SThreadCtxt_t *)pvArg;
    boolean_t      bContinue = TRUE;
    boolean_t      bPrint = TRUE;

    daemonize("flow_thread%d", pContext->m_iIndex);

    allow_signal(SIGKILL);

    pContext->m_pTask = current;
    complete(&pContext->m_sStartComp);

    while(bContinue)
    {
        allow_signal(SIGHUP);

        set_current_state(TASK_INTERRUPTIBLE);
        schedule();

        while(bContinue && signal_pending(current))
        {
            unsigned long ulSignal;
            siginfo_t     sigInfo;

            ulSignal =
                dequeue_signal_lock(current, &current->blocked, &sigInfo);

            switch(ulSignal)
            {
                case SIGKILL:
                    bContinue = FALSE;
                    break;
                default:
                    break;
            }
        }

        if(bContinue)
        {
            disallow_signal(SIGHUP);

            while(FLOW_Q_TYPE_EMPTY !=
                  pContext->m_asQ[pContext->m_iRIndex].m_eType)
            {
                SHashLevel_t *pHashLevel = NULL;
                SQEntry_t    *psEntry = &pContext->m_asQ[pContext->m_iRIndex];
                SFlow_t      *psFlow = NULL;
                u32           u32Index = 0;
                
                if(FLOW_Q_TYPE_DEL != psEntry->m_eType &&
                   NULL == psEntry->skb)
                {
                    continue;
                }

                switch(psEntry->m_eType)
                {
                    case FLOW_Q_TYPE_EXCEPTION:
                        ProcessException(psEntry->skb, &psEntry->m_uTupple);
                        break;
                    case FLOW_Q_TYPE_DEL:
                        ProccessDel(psEntry->m_pNetDev, &psEntry->m_uTupple);
                        break;
                    case FLOW_Q_TYPE_NORMAL:
                        if(FlowExists(psEntry->skb,&psEntry->m_uTupple,
                                      &psFlow, &pHashLevel, &u32Index) ||
                           FindFlow(psEntry->skb, &psEntry->m_uTupple,
                                    &psFlow, pHashLevel, u32Index, pContext))
                        {
                            ProcessFlow(psEntry->skb, psFlow,
                                        &psEntry->m_uTupple, TRUE);
                        }
                        else
                        {
                            if(bPrint)
                            {
                                printk(KERN_ERR "%d: Could not find flow!\n",
                                       pContext->m_iIndex);
                                bPrint = FALSE;
                            }
                            dev_kfree_skb_any(psEntry->skb);
                        }
                        break;
                    default:
                        if(psEntry->skb)
                        {
                            dev_kfree_skb_any(psEntry->skb);
                        }
                        break;
                }
                psEntry->skb = NULL;
                psEntry->m_eType = FLOW_Q_TYPE_EMPTY;
                pContext->m_iRIndex = (pContext->m_iRIndex + 1) % FLOW_Q_DEPTH;
            }
        }
    }

    pContext->m_pTask = NULL;

    // Cleanup
    if(pContext->m_asQ[pContext->m_iRIndex].skb)
    {
        printk(KERN_WARNING "Q not empty!\n");
    }

    while(pContext->m_asQ[pContext->m_iRIndex].skb)
    {
        dev_kfree_skb_any(pContext->m_asQ[pContext->m_iRIndex].skb);
        pContext->m_asQ[pContext->m_iRIndex].skb = NULL;
        pContext->m_iRIndex = (pContext->m_iRIndex + 1) % FLOW_Q_DEPTH;
    }

    complete(&pContext->m_sExitComp);

    return 0;
}

static enum
{
    FLOWDEV_NONE = 0,
    FLOWDEV_READ,
    FLOWDEV_WRITE,
} sg_eWhat = FLOWDEV_NONE;
static boolean_t sg_bInUse = FALSE;
void FreeXferBuffer(void);

ssize_t flow_dev_read(struct file *filp, char *buff, size_t count,
                      loff_t *offp)
{
    ssize_t sRet = 0;

    if(FLOWDEV_NONE == sg_eWhat)
    {
        sg_eWhat = FLOWDEV_READ;
    }
    else if(FLOWDEV_READ != sg_eWhat)
    {
        return -EBUSY;
    }

    if(*offp > gs_u32XferBufSize)
    {
        return -EINVAL;
    }

    if(gs_u32XferBufSize && gs_pu8XferBuffer)
    {
        sRet = count;

        if((*offp + count) > gs_u32XferBufSize)
        {
            sRet = gs_u32XferBufSize - *offp;
        }

        if(sRet && copy_to_user(buff, &gs_pu8XferBuffer[*offp], sRet))
        {
            return -EFAULT;
        }

        *offp += sRet;
    }

    return sRet;
}

ssize_t flow_dev_write(struct file *filp, const char *buff, size_t count,
                       loff_t *offp)
{
    if(FLOWDEV_NONE == sg_eWhat)
    {
        sg_eWhat = FLOWDEV_WRITE;
        FreeXferBuffer();
    }
    else if(FLOWDEV_WRITE != sg_eWhat)
    {
        return -EBUSY;
    }

    if(NULL == gs_pu8XferBuffer)
    {
        if(*offp || count < 4)
        {
            return -EINVAL;
        }

        if(copy_from_user(&gs_u32XferBufSize, buff, sizeof(gs_u32XferBufSize)))
        {
            return -EFAULT;
        }

        if(!(gs_pu8XferBuffer = (u8 *)vmalloc(gs_u32XferBufSize)))
        {
            gs_u32XferBufSize = 0;
            return -ENOMEM;
        }
    }

    if(copy_from_user(&gs_pu8XferBuffer[*offp], buff, count))
    {
        return -EFAULT;
    }

    *offp += count;

    return count;
}

void FreeXferBuffer(void)
{
    if(gs_pu8XferBuffer)
    {
        dtn_node  *tmpdnode = NULL;
        dtn_node  *dnode = glbl_root_dtn_node;

        while(dnode)
        {
            if(!dnode->temp_index && dnode->r_list)
            {
                kfree(dnode->r_list);
                dnode->r_list = NULL;
            }

            if(dnode->temp_index < dnode->sub_tree_size)
            {
                tmpdnode = dnode->sub_tree[dnode->temp_index++];
                tmpdnode->temp_index = 0;
                dnode = tmpdnode;
            }
            else
            {
                dnode = dnode->parent;
            }
        }

        vfree(gs_pu8XferBuffer);
    }

    gs_pu8XferBuffer = NULL;
    gs_u32XferBufSize = 0;
    glbl_root_dtn_node = NULL;
}

// #define DEBUG_XFER
#ifdef DEBUG_XFER
#define MY_PRINT(format,__VARGS__...) printk(KERN_INFO format, ##__VARGS__);
#include"../../HiCuts/obj_print.inc"
#endif /* DEBUG_XFER */

void MyPrint(void)
{
    dtn_node  *dnode = glbl_root_dtn_node;
    dtn_node  *tmpdnode = NULL;
    int        iNumNodes = 0;
    int        iNumRules = 0;
    int        iLevel = 0;

    printk(KERN_INFO "%s: %d bytes (@ %p)\n", __FUNCTION__, gs_u32XferBufSize,
           gs_pu8XferBuffer);

    dnode->temp_index = 0;

    while(dnode)
    {
        if(!dnode->temp_index)
        {
#ifdef DEBUG_XFER
            print_dtn_node_info(dnode, iLevel);
#endif /* DEBUG_XFER */
            iNumNodes++;
            iNumRules += dnode->num_of_rules;
        }

        if(dnode->temp_index < dnode->sub_tree_size)
        {
            iLevel++;
            tmpdnode = dnode->sub_tree[dnode->temp_index++];
            tmpdnode->temp_index = 0;
            dnode = tmpdnode;
        }
        else
        {
            dnode = dnode->parent;
            iLevel--;
        }
    }

    printk(KERN_INFO "%d nodes, %d rule indexes\n", iNumNodes, iNumRules);
}

void LinkDecisionTree(void)
{
    dtn_node *pnode = NULL;
    dtn_node *dnode = NULL;
    size_t    sz = 0;
    rule     *pRules = NULL;
    u32      *pRIndexes = NULL;
    u32       u32NumRules = 0;
    int       iIndex = 0;
    int       ii;

	printk(KERN_WARNING "Entered LinkDecisionTree...\n");
    // Make sure things make sense
    if(NULL == gs_pu8XferBuffer || gs_u32XferBufSize < (2 * sizeof(u32)))
    {
        return;
    }

    // get the number of rules
    iIndex += sizeof(u32);
    u32NumRules = *((u32 *) &gs_pu8XferBuffer[iIndex]);

	printk(KERN_WARNING "number of rules: %d\n", u32NumRules);
    // get the start of the rules array
    iIndex += sizeof(u32);
    pRules = (rule *) &gs_pu8XferBuffer[iIndex];

    // Make sure that there's enough data to hold the rules
    iIndex += u32NumRules * sizeof(rule);
    if(iIndex >= gs_u32XferBufSize)
    {
		printk(KERN_WARNING "Not enough data to hold rules\n");
        return;
    }

    for(;iIndex < gs_u32XferBufSize;)
    {
        // Make sure that there's at least enough data for the node
        dnode = (dtn_node *)&gs_pu8XferBuffer[iIndex];
        iIndex += sizeof(dtn_node);
        if(iIndex >= gs_u32XferBufSize)
        {
			printk(KERN_WARNING "Not enough data for the node\n");
            return;
        }
        dnode->temp_index = 0;

        // Make sure that there's enough data for the rule indexes
        pRIndexes = (u32 *) &gs_pu8XferBuffer[iIndex];
        iIndex += dnode->num_of_rules * sizeof(u32);
        if(iIndex > gs_u32XferBufSize)
        {
			printk(KERN_WARNING "not enough data for the rule indexes\n");
            return;
        }

        // Allocate space for the rules and add them
        if(dnode->num_of_rules)
        {
            sz = dnode->num_of_rules * sizeof(rule *);
            if((dnode->r_list = (rule **) kmalloc(sz, GFP_KERNEL)))
            {
                memset(dnode->r_list, 0x00, sz);
                for(ii = 0;ii < dnode->num_of_rules;ii++)
                {
                    if(pRIndexes[ii] < u32NumRules)
                    {
                        dnode->r_list[ii] = &pRules[pRIndexes[ii]];
                    }
                }
            }
            else
            {
                printk(KERN_ERR "%s:%d Unable to allocate memory!\n",
                       __FUNCTION__, __LINE__);
                return;
            }
        }

        // Link it to the tree
        if(NULL == glbl_root_dtn_node)
        {
            glbl_root_dtn_node = dnode;
        }

        if((dnode->parent = pnode))
        {
            pnode->sub_tree[pnode->temp_index++] = dnode;
        }

        // Allocate space for any subtrees
        if(dnode->sub_tree_size)
        {
            sz = dnode->sub_tree_size * sizeof(dtn_node *);
            if((dnode->sub_tree = (dtn_node **) kmalloc(sz, GFP_KERNEL)))
            {
                memset(dnode->sub_tree, 0x00, sz);
            }
            else
            {
                printk(KERN_ERR "%s:%d Unable to allocate memory!\n",
                       __FUNCTION__, __LINE__);
                return;
            }
        }

        if(dnode->sub_tree_size)
        {
            pnode = dnode;
        }
        else if(pnode)
        {
            while(pnode && (pnode->temp_index >= pnode->sub_tree_size))
            {
                pnode = pnode->parent;
            }
        }

        dnode = NULL;
    }

    MyPrint();
}

int flow_dev_open(struct inode *inode, struct file *filp)
{
    if(sg_bInUse)
    {
        return -EBUSY;
    }

    sg_bInUse = TRUE;

    return 0;
}

int flow_dev_release(struct inode *inode, struct file *filp)
{
    if(FLOWDEV_WRITE == sg_eWhat)
    {
        LinkDecisionTree();
    }
    sg_eWhat  = FLOWDEV_NONE;
    sg_bInUse = FALSE;
    return 0;
}

#define FLOW_DEV_NAME "flow_dev"
int flow_dev_major = -1;

struct file_operations flow_dev_fops = {
    read:    flow_dev_read,
    write:   flow_dev_write,
    open:    flow_dev_open,
    release: flow_dev_release,
};

int CreateDevNode()
{
    if((flow_dev_major =
        register_chrdev(0, FLOW_DEV_NAME, &flow_dev_fops)) < 0)
    {
        printk(KERN_ERR "Could not create " FLOW_DEV_NAME " device\n");
    }

    return flow_dev_major;
}

void DestroyDevNode()
{
    if(flow_dev_major >= 0)
    {
        unregister_chrdev(flow_dev_major, FLOW_DEV_NAME);
        FreeXferBuffer();
    }
}

#if 0
__be32 htonl(__u32 arg)
{
    return (( __be32)(__builtin_constant_p((__u32)((arg))) ? ({ __u32 __x = (((arg))); ((__u32)( (((__u32)(__x) & (__u32)0x000000ffUL) << 24) | (((__u32)(__x) & (__u32)0x0000ff00UL) << 8) | (((__u32)(__x) & (__u32)0x00ff0000UL) >> 8) | (((__u32)(__x) & (__u32)0xff000000UL) >> 24) )); }) : __fswab32(((arg)))));
}

__u16 ntohs(__be16 arg)
{
    return (__builtin_constant_p((__u16)(( __u16)(__be16)(arg))) ? ({ __u16 __x = ((( __u16)(__be16)(arg))); ((__u16)( (((__u16)(__x) & (__u16)0x00ffU) << 8) | (((__u16)(__x) & (__u16)0xff00U) >> 8) )); }) : __fswab16((( __u16)(__be16)(arg))));
}
#endif
